"""Mutation operators for DepotPrototype."""
import random
from eflips.depot.layout_opt.opt_tools import init

mutations = []  # for logging


def mutgaussian_area_count_crop(depot, mu, sigma):
    """Random gaussian mutation of area count.
    If less than before, the area list is cropped on the right. If higher than
    before, new areas with random init values are appended to the right.
    """
    diff = round(random.gauss(mu, sigma))
    count_old = len(depot.areas)
    count_new = count_old + diff

    if count_new < init.COUNT_MIN or count_new > init.COUNT_MAX:
        # boundary violation
        return False
    if diff == 0:
        # no need to take action
        return False

    if diff < 0:
        depot.areas = depot.areas[:count_new]
    elif diff > 0:
        new_areas = [init.init_random_area() for _ in range(diff)]
        depot.areas.extend(new_areas)

    assert len(depot.areas) == count_new

    return True


def mutgaussian_area_count_rnd(depot, mu, sigma):
    """Random gaussian mutation of area count.
    If less than before, random areas are removed. If higher than before, new
    areas with random capacities are appended to the right.
    """
    diff = round(random.gauss(mu, sigma))
    count_old = len(depot.areas)
    count_new = count_old + diff

    if count_new < init.COUNT_MIN or count_new > init.COUNT_MAX:
        # boundary violation
        return False
    if diff == 0:
        # no need to take action
        return False

    if diff < 0:
        removals = random.sample(depot.areas, abs(diff))
        for removal in removals:
            depot.areas.remove(removal)
    elif diff > 0:
        new_areas = [init.init_random_area() for _ in range(diff)]
        depot.areas.extend(new_areas)

    assert len(depot.areas) == count_new

    if diff > 0:
        mutations.append("count+")
    else:
        mutations.append("count-")

    return True


def mutuniform_area_count_crop(depot, a, b):
    """Random uniform mutation of area count.
    If less than before, the area list is cropped on the right. If higher than
    before, new areas with random init values are appended to the right.
    """
    if a < 0:
        raise ValueError("a must be > 0.")
    if a > b:
        raise ValueError("a must be < b")

    diff = random.randint(a, b)
    count_old = len(depot.areas)

    if random.random() < 0.5:
        diff = -diff

    count_new = count_old + diff

    if count_new < init.COUNT_MIN or count_new > init.COUNT_MAX:
        # boundary violation
        return False
    if diff == 0:
        # no need to take action
        return False

    if diff < 0:
        depot.areas = depot.areas[:count_new]
    elif diff > 0:
        new_areas = [init.init_random_area() for _ in range(diff)]
        depot.areas.extend(new_areas)

    assert len(depot.areas) == count_new

    return True


def mutuniform_area_count_rnd(depot, a, b):
    """Random uniform mutation of area count.
    If less than before, the area list is cropped on the right. If higher than
    before, new areas with random init values are appended to the right.
    """
    if a < 0:
        raise ValueError("a must be > 0.")
    if a > b:
        raise ValueError("a must be < b")

    diff = random.randint(a, b)
    count_old = len(depot.areas)

    if random.random() < 0.5:
        diff = -diff

    count_new = count_old + diff

    if count_new < init.COUNT_MIN or count_new > init.COUNT_MAX:
        # boundary violation
        return False
    if diff == 0:
        # no need to take action
        return False

    if diff < 0:
        removals = random.sample(depot.areas, abs(diff))
        for removal in removals:
            depot.areas.remove(removal)
    elif diff > 0:
        new_areas = [init.init_random_area() for _ in range(diff)]
        depot.areas.extend(new_areas)

    assert len(depot.areas) == count_new

    return True


def mutgaussian_area_capacity(area, mu, sigma):
    """Random gaussian mutation of area capacity."""
    capacity_old = area.capacity
    capacity_new = capacity_old + round(random.gauss(mu, sigma))

    if capacity_new < area.capacity_min or capacity_new > area.capacity_max:
        # boundary violation
        return False
    if capacity_new == capacity_old:
        # no need to take action
        return False

    area.capacity = capacity_new
    area.visu = None

    if capacity_new > capacity_old:
        mutations.append("capacity+")
    else:
        mutations.append("capacity-")

    return True


def mutuniform_area_capacity(area, a, b):
    """Random uniform mutation of area capacity."""
    if a < 0:
        raise ValueError("a must be > 0.")
    if a > b:
        raise ValueError("a must be < b")

    diff = random.randint(a, b)
    capacity_old = area.capacity

    if random.random() < 0.5:
        capacity_new = capacity_old + diff
    else:
        capacity_new = capacity_old - diff

    if capacity_new < area.capacity_min or capacity_new > area.capacity_max:
        # boundary violation
        return False
    if capacity_new == capacity_old:
        # no need to take action
        return False

    area.capacity = capacity_new
    area.visu = None
    return True


def mut_area_to_other_type(area, depot):
    """Select a random area and convert it to one or more areas of another
    type, keeping the capacity."""
    this_type = type(area)
    other_type = random.choice(init.OTHER_AREA_TYPES[this_type])
    new_areas = area.to_other_type(other_type)

    if not new_areas:
        return False

    depot.areas.remove(area)
    # Append new areas to the end. The change of order is okay because
    # areas are sorted before packing.
    depot.areas.extend(new_areas)
    mutations.append("type_to_other")
    return True


def mut_combine_two_areas(depot):
    """Select two random areas and try to combine them."""
    if len(depot.areas) < 2:
        # Not enough areas
        # print('Not enough areas.')
        return False

    a1, a2 = random.sample(depot.areas, 2)
    # print('Attempting to combine %s into %s' % (a2, a1))

    if a1.capacity + a2.capacity > a1.capacity_max:
        # print('Combination of %s into %s would violate the capacity max of %d (wanted %d).' % (a2.typename, a1.typename, a1.capacity_max, a1.capacity + a2.capacity))
        return False

    a1.capacity += a2.capacity
    a1.visu = None
    depot.areas.remove(a2)
    # print('Successfully combined into %s.' % a1)
    mutations.append("type_combine")
    return True


def mutgaussian_depot_and(
    depot,
    mutsigma_area_count,
    mutpb_area_count,
    mutpb_area_type,
    mutsigma_area_capacity,
    mutpb_area_capacity,
):
    """Wrapper for mutations of a depot individual where area count, type and
    capacity might be mutated.
    """
    mutated = False
    if random.random() < mutpb_area_count:
        mutated = mutgaussian_area_count_rnd(depot, 0, mutsigma_area_count) or mutated

    for area in depot.areas.copy():
        if random.random() < mutpb_area_capacity:
            mutated = (
                mutgaussian_area_capacity(area, 0, mutsigma_area_capacity) or mutated
            )

        if random.random() < mutpb_area_type:
            mutated = mut_area_to_other_type(area, depot) or mutated

    if mutated:
        depot.reset_results()

    return mutated


def mutgaussian_depot_or_partial(
    depot,
    mutsigma_area_count,
    mutpb_area_count,
    mutpb_area_type,
    mutsigma_area_capacity,
    mutpb_area_capacity,
):
    """Wrapper for mutations of a depot individual where area count and either
    type or capacity might be mutated.
    """
    if mutpb_area_capacity + mutpb_area_type > 1:
        raise ValueError(
            "The sum of mutpb_area_type and mutpb_area_capacity "
            "must be smaller or equal to 1."
        )

    mutated = False
    if random.random() < mutpb_area_count:
        mutated = mutgaussian_area_count_rnd(depot, 0, mutsigma_area_count) or mutated

    for area in depot.areas.copy():
        mut_choice = random.random()

        if mut_choice < mutpb_area_capacity:
            mutated = (
                mutgaussian_area_capacity(area, 0, mutsigma_area_capacity) or mutated
            )

        elif mut_choice < mutpb_area_capacity + mutpb_area_type:
            mutated = mut_area_to_other_type(area, depot) or mutated

    if mutated:
        depot.reset_results()

    return mutated


def mutgaussian_depot_or(
    depot,
    mutsigma_area_count,
    mutpb_area_count,
    mutpb_area_type,
    mutsigma_area_capacity,
    mutpb_area_capacity,
    splitpb,
):
    """Wrapper for mutations of a depot individual where either area count or
    type or capacity of one area might be mutated.
    """
    if mutpb_area_count + mutpb_area_capacity + mutpb_area_type > 1:
        raise ValueError(
            "The sum of mutsigma_area_count, mutpb_area_type and "
            "mutpb_area_capacity must be smaller or equal to 1."
        )

    mutated = False
    mut_choice = random.random()

    if mut_choice < mutpb_area_count:
        mutated = mutgaussian_area_count_rnd(depot, 0, mutsigma_area_count) or mutated

    elif mut_choice < mutpb_area_count + mutpb_area_capacity:
        area = random.choice(depot.areas)
        mutated = mutgaussian_area_capacity(area, 0, mutsigma_area_capacity) or mutated

    elif mut_choice < mutpb_area_count + mutpb_area_capacity + mutpb_area_type:
        if random.random() < splitpb:  # p_split
            # Mutate one area to another type, possibly splitting it up
            area = random.choice(depot.areas)
            mutated = mut_area_to_other_type(area, depot) or mutated
        else:
            # Combine two areas
            mutated = mut_combine_two_areas(depot)

    if mutated:
        depot.reset_results()

    return mutated


def mutuniform_depot_or_partial(
    depot,
    a_area_count,
    b_a_area_count,
    mutpb_area_count,
    mutpb_area_type,
    a_area_capacity,
    b_area_capacity,
    mutpb_area_capacity,
):
    """Wrapper for mutations of a depot individual where area count and either
    type or capacity might be mutated.
    """
    if mutpb_area_capacity + mutpb_area_type > 1:
        raise ValueError(
            "The sum of mutpb_area_type and mutpb_area_capacity "
            "must be smaller or equal to 1."
        )

    mutated = False
    if random.random() < mutpb_area_count:
        mutated = (
            mutuniform_area_count_crop(depot, a_area_count, b_a_area_count) or mutated
        )

    for area in depot.areas.copy():
        mut_choice = random.random()

        if mut_choice < mutpb_area_capacity:
            mutated = (
                mutuniform_area_capacity(area, a_area_capacity, b_area_capacity)
                or mutated
            )

        elif mut_choice < mutpb_area_capacity + mutpb_area_type:
            mutated = mut_area_to_other_type(area, depot) or mutated

    if mutated:
        depot.reset_results()

    return mutated


def mutuniform_depot_or(
    depot,
    a_area_count,
    b_a_area_count,
    mutpb_area_count,
    mutpb_area_type,
    a_area_capacity,
    b_area_capacity,
    mutpb_area_capacity,
):
    """Wrapper for mutations of a depot individual where area count and either
    type or capacity might be mutated.
    """
    if mutpb_area_count + mutpb_area_capacity + mutpb_area_type > 1:
        raise ValueError(
            "The sum of mutsigma_area_count, mutpb_area_type and "
            "mutpb_area_capacity must be smaller or equal to 1."
        )

    mutated = False
    mut_choice = random.random()

    if mut_choice < mutpb_area_count:
        mutated = (
            mutuniform_area_count_rnd(depot, a_area_count, b_a_area_count) or mutated
        )

    elif mut_choice < mutpb_area_count + mutpb_area_capacity:
        area = random.choice(depot.areas)
        mutated = (
            mutuniform_area_capacity(area, a_area_capacity, b_area_capacity) or mutated
        )

    elif mut_choice < mutpb_area_count + mutpb_area_capacity + mutpb_area_type:
        area = random.choice(depot.areas)
        mutated = mut_area_to_other_type(area, depot) or mutated

    if mutated:
        depot.reset_results()

    return mutated



# -*- coding: utf-8 -*-
import warnings

import eflips
from eflips.settings import globalConstants
from eflips.helperFunctions import flexprint


class VehicleFilter:
    """Class to build and apply filters for vehicles. Designed to work with
    simpy FilterStore and subclasses in the depot simulation: Implements
    __call__ to make an instance directly callable and return the all()-result
    of the criteria. Can be recalled. Usable e.g. for simpyExt.FilterStoreExt
    that needs the store item as single positional argument for the filter
    function. Filter criteria are prepared in class methods and can be selected
    and combined before application.

    Parameters:
    filter_names: [list or None] of filter method names (as strings, 'filter_'
        omitted) of this class that should be applied on a vehicle. If left
        None, the VehicleFilter returns True.
    All other parameters that are required by selected filter methods must be
        added as keyword arguments during init (see method descriptions).

    Attributes:
    filters: [list] of one or more VehicleFilter.filter_... methods or other
        valid callables. Filled automatically from *filter_names*. Can be None
        or empty, in which case the VehicleFilter returns True. Entries can be
        added during definition (see filter_names) and afterwards (see
        VehicleFilter.append).

    Usage:
        Instantiate a VehicleFilter object. Pass arguments required for the
        checks, such as filter_names upon instantiation as kwargs or add them
        later.
        Example:
            vf = VehicleFilter(filter_names=['service_need'], env=env)
            trueorfalse = vf(vehicle)
        Static methods may be called without instantiation. Example:
            trueorfalse = VehicleFilter.filter_no_active_processes(vehicle)

    When adding a new filter method:
    - The method name must be preceeded by 'filter_'.
    - 'vehicle' must be the only method parameter. Additional parameters can be
        accessed via the VehicleFilter object attributes. Passing all those
        parameters as keyword arguments during init adds them automatically as
        attribute of the VehicleFilter. Alternatively, attributes might be
        added after init, before calling the filter.

    """

    def __init__(self, filter_names=None, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.filter_names = filter_names
        if self.filter_names is None:
            self.filter_names = []
        elif not isinstance(filter_names, list):
            raise ValueError(
                "VehicleFilter 'filter_names' must be of type "
                "list, not %s" % type(filter_names)
            )

        self.vehicle_types_converted = False

        if hasattr(self, "period"):
            if (
                not isinstance(self.period, tuple)
                and not isinstance(self.period, list)
                or len(self.period) != 2
                or not 0 <= self.period[0] <= 86400
                or not 0 <= self.period[1] <= 86400
                or self.period[0] == self.period[1]
            ):
                raise ValueError(
                    "Invalid VehicleFilter.period. See "
                    "description of method filter_in_period for "
                    "instructions."
                )

        self.filters = []
        self.filters_from_names()

    def __call__(self, vehicle):
        """Method that actually applies filters. Return True if all of
        self.filter return True or self.filters is empty."""
        result = all(filter(vehicle) for filter in self.filters)
        return result

    def any(self, vehicle):
        """Return True if any of the filters in self.filters is True or
        self.filters is empty. Optional method in addition to the standard
        usage with logical all which is implemented in __call__."""
        result = any(filter(vehicle) for filter in self.filters)
        return result

    def filters_from_names(self):
        """Append references to methods specified in filter_names to
        self.filters. Extends the given strings by 'filter_' at the beginning,
        so that part does not have to be passed during instantiation.
        Filters can alternatively be added after instantiation using append().
        """
        for filter_name in self.filter_names:
            filter_name = "filter_" + filter_name
            if not hasattr(self, filter_name):
                raise ValueError(
                    'Method "%s" not found in class VehicleFilter' % filter_name
                )
            filter = getattr(self, filter_name)
            self.append(filter)

    def append(self, filter):
        """Append *filter* to self.filters. Filter must be a callable accepting
        vehicle as argument and returning a boolean value."""
        self.filters.append(filter)

    def filter_vehicle_type(self, vehicle):
        """Return True if the vehicle type matches ANY item in
        self.vehicle_types.

        Attributes required in self:
        vehicle_types: [list] of eflips.depot.simple_vehicle.VehicleType
            objects
        """
        self.get_vt_objects()
        return vehicle.vehicle_type in self.vehicle_types

    def get_vt_objects(self, force=False):
        """Helper for filter methods using self.vehicle_types. Get objects for
        strings in self.vehicle_types.
        If *force* is False and self.vehicle_types_converted is True, the
        conversion is skipped (useful for reusing the filter object with the
        same parameters). If *force* is true, the conversion is done even if
        self.vehicle_types_converted is True (useful for reusing the filter
        object with new parameters).

        Attributes required in self:
        vehicle_types: [list] of vehicle type IDs as strings or
            eflips.depot.simple_vehicle.VehicleType objects
        """
        if not self.vehicle_types_converted or force:
            # Get obj references of vehicle types
            try:
                self.vehicle_types_str = (
                    self.vehicle_types.copy()
                )  # backup the list of strings
            except AttributeError:
                raise AttributeError(
                    "filter_vehicle_type requires the "
                    "attribute vehicle_types to be set before call."
                )
            # print(self.vehicle_types, self.vehicle_types_str)
            for no, ID in enumerate(self.vehicle_types):
                if isinstance(ID, str):
                    try:
                        self.vehicle_types[no] = globalConstants["depot"][
                            "vehicle_types_obj_dict"
                        ][ID]
                    except KeyError:
                        raise KeyError(
                            "Vehicle type %s not found in globalConstants. "
                            "Possibly selected eflips settings and depot template "
                            "don't match." % ID
                        )
            self.vehicle_types_converted = True

    def filter_trip_vehicle_match(self, vehicle):
        """Return True if vehicle is suitable for self.trip (for vehicles that
        have not entered the system yet) or if vehicle and self.trip were
        matched by the dispatching.

        Attributes required in self:
        trip: eflips.depot.standalone.SimpleTrip object
        """
        return (
            not vehicle.system_entry
            and vehicle.vehicle_type in self.trip.vehicle_types
            or vehicle.trip is not None
            and vehicle.trip is self.trip
        )

    @staticmethod
    def filter_no_active_processes(vehicle):
        """Return True if the vehicle has no active processes."""
        return not vehicle.dwd.any_active_processes

    @staticmethod
    def filter_no_active_uncancellable_processes(vehicle):
        """Return True if the vehicle has no active processes or only those
        that can be cancelled for dispatch.
        """
        if not vehicle.dwd.any_active_processes:
            return True
        elif not vehicle.dwd.active_processes:
            # If active_processes is empty but any_active_processes is True, a
            # process might be about to start but is not listed in
            # active_processes yet. The check may return True one second later
            # if the process is cancellable.
            return False
        else:
            return all(
                proc.cancellable_for_dispatch for proc in vehicle.dwd.active_processes
            )

    @staticmethod
    def filter_bat_full(vehicle):
        """Return True if the vehicle's battery is full."""
        return vehicle.battery.soc == 1

    def filter_min_energy(self, vehicle):
        """Return True if the remaining energy of the vehicle's battery is at
        least self.min_energy.

        Attributes required in self:
        min_energy: [int] minimum battery level
        """
        return self.min_energy <= vehicle.battery.energy_remaining

    def filter_sufficient_energy(self, vehicle):
        """Return True if the remaining energy of the vehicle's battery plus a
        reserve is sufficient for the trip.

        Attributes required in self:
        trip: eflips.depot.standalone.SimpleTrip object
        """
        if globalConstants["depot"]["consumption_calc_mode"] == "CR_distance_based":
            energy_reserve = globalConstants["depot"]["energy_reserve"]
            required_energy = (
                vehicle.vehicle_type.CR
                * self.trip.distance
                * (1 + (energy_reserve / 100))
            )
            result = required_energy <= vehicle.battery.energy_remaining

        elif globalConstants["depot"]["consumption_calc_mode"] == "CR_time_based":
            energy_reserve = globalConstants["depot"]["energy_reserve"]
            required_energy = (
                self.trip.duration
                / 3600
                * vehicle.vehicle_type.CR
                * (1 + (energy_reserve / 100))
            )
            result = required_energy <= vehicle.battery.energy_remaining

        elif globalConstants["depot"]["consumption_calc_mode"] == "soc_given":
            if self.trip.charge_on_track:
                result = round(vehicle.battery.soc, 5) >= self.trip.start_soc
            else:
                required_energy = (
                    self.trip.start_soc - self.trip.end_soc
                ) * vehicle.battery.energy_real

                # If the vehicle is fully charged and its fully charged energy is still lower than the required energy,
                # dispatch anyway and warn the user
                if (
                    abs(vehicle.battery.soc - 1) < 1e-6
                    and vehicle.battery.energy_real < required_energy
                ):
                    warnings.warn(
                        f"Vehicle {vehicle.ID} is fully charged but the required energy for the trip is higher than the fully charged energy. Dispatching anyway."
                    )
                    return True

                result = required_energy <= vehicle.battery.energy_remaining

        else:
            raise ValueError(
                "Invalid value %s for 'consumption_calc_mode' in globalConstants."
                % globalConstants["depot"]["consumption_calc_mode"]
            )

        # flexprint(
        #     '\tVehicle %s in battery level check. result: %s. Battery energy_remaining: %d (SoC=%.3f). Trip.energy need: %d'
        #     % (vehicle.ID, result, vehicle.battery.energy_remaining, vehicle.battery.soc,
        #        required_energy), switch='departure_before_fully_charged')
        return result

    def filter_soc_lower_than(self, vehicle):
        """Return True if the soc of the vehicle's battery is lower than
        self.soc.

        Attributes required in self:
        soc: [float]
        """
        return vehicle.battery.soc < self.soc

    def filter_service_need(self, vehicle):
        """Return True if the vehicle needs service, i.e. the cumulated trip
        time since last service is greater than self.service_need_td or the
        elapsed time since last service is greater than
        self.service_need_t_elapsed.
        This filter is one option to check the service need. An independent
        alternative is filter_in_period.

        Attributes required in self:
        env: simpy.Environment object
        service_need_td: [int] amount of driving time [seconds] since last
          service from which service_need will be set to True when exceeded.
        service_need_t_elapsed: [int] amount of elapsed time [seconds] since
            last service from which service_need will be set to True when
            exceeded.
        """
        # Skip checks if service_need already is True
        if not vehicle.dwd.service_need:
            t_trips = 0
            # Get cumulated trip time since last service
            for trip in reversed(vehicle.finished_trips):
                if trip.atd > vehicle.dwd.t_lastServed:
                    t_trips += trip.ata - trip.atd
                else:
                    break
            if (
                t_trips > self.service_need_td
                or self.env.now - vehicle.dwd.t_lastServed > self.service_need_t_elapsed
            ):
                vehicle.dwd.service_need = True
        return vehicle.dwd.service_need

    @staticmethod
    def filter_repair_need(vehicle):
        """Return True if the vehicle needs repair."""
        return vehicle.dwd.repair_need

    @staticmethod
    def filter_maintenance_need(vehicle):
        """Return True if the vehicle needs maintenance."""
        return vehicle.dwd.maintenance_need

    def filter_in_period(self, vehicle):
        """Return True if self.now as a daytime is in self.period.

        Attributes required in self:
        env: simpy.Environment object
        period: [tuple or list] daytime period (t0, t1) for
            filter_in_period. t0 and t1 are seconds with 0 < t < 86400. t1 may
            be lower than t0 to specify periods including midnight. t0 and t1
            cannot be equal. (reason for list is enabling direct json input)
        """
        t0 = self.period[0]
        t1 = self.period[1]
        now = self.env.now % 86400  # get daytime in seconds
        if t0 < t1:
            return t0 <= now <= t1
        else:
            return t0 <= now <= 86400 or now <= t1

    def filter_in_period_days(self, vehicle):
        """Return True if the time is in period and it is the third, fourth,
        ... day.

        Attributes required in self:
        env: simpy.Environment object
        after_day: [int] determines in how many days a service has to be
            done. For example every third day.
        """
        id = sum(
            [int(id) for id in vehicle.ID.split() if id.isdigit()]
        )  # transfers the str ID in a numeric id
        day = int(self.env.now / 86400) + 1
        if id % self.after_day == day % self.after_day and self.filter_in_period(
            vehicle
        ):
            return True

    @staticmethod
    def filter_isunblocked(vehicle):
        if isinstance(vehicle.dwd.current_area, eflips.depot.DirectArea):
            result = True
        else:
            result = vehicle.dwd.current_area.isunblocked(vehicle)
        return result

    @staticmethod
    def filter_not_on_hold(vehicle):
        """Return True if *vehicle* is not marked to be on hold."""
        return not vehicle.dwd.on_hold

    def filter_dwd_previous_area(self, vehicle):
        """Check if current area is in none_of_previous_areas.

        Attributes required in self:
        none_of_previous_areas: [list] Skip current area if previous
            area in none_of_previous_areas.
        """
        return vehicle.dwd.current_area.ID not in self.none_of_previous_areas

    @staticmethod
    def filter_false(vehicle):
        """Always return False."""
        return False



# -*- coding: utf-8 -*-
"""
Resources for the depot simulation.

"""
import itertools
import simpy
from simpy.core import BoundClass
from simpy.resources.resource import PriorityRequest, Release
from eflips.settings import globalConstants
from eflips.evaluation import DataLogger
from eflips.helperFunctions import flexprint
from eflips.depot.processes import BaseDepotProcess


class DepotResourceRequest(PriorityRequest):
    """Request event for DepotResource. Subclass of simpy.resources.resource.
    PriorityRequest, for providing an interface for customization.

    caller: object that issued the request. Must have attribute 'ID'
        (e.g. 'serve').
    preempt: [bool] switch for preemption. Same behaviour as in class
        PriorityRequest. Only set to False if it is certain that this request
        is not used with simpy.AllOf alongside other requests to avoid a
        potential locked resource case (currently done so in
        DepotControl.run_processes). See depot_simulation_tests_depotprocess.py
        for explanation and demo.

    """

    def __init__(self, resource, caller, priority=0, preempt=True):
        super(DepotResourceRequest, self).__init__(resource, priority, preempt)

        self.caller = caller

        if (
            hasattr(resource, "logger")
            and globalConstants["general"]["LOG_ATTRIBUTES"]
            and globalConstants["general"]["LOG_SPECIFIC_STEPS"]
        ):
            self.callbacks.append(resource.logger.steplog)


class DepotResourceRelease(Release):
    """Release event for DepotResource. Subclass of simpy.resources.resource.
    Release, for providing an interface for customization.

    """

    def __init__(self, resource, request):
        super(DepotResourceRelease, self).__init__(resource, request)

        if (
            hasattr(resource, "logger")
            and globalConstants["general"]["LOG_ATTRIBUTES"]
            and globalConstants["general"]["LOG_SPECIFIC_STEPS"]
        ):
            self.callbacks.append(resource.logger.steplog)


class DepotResource(simpy.PreemptiveResource):
    """Extension of simpy PreemptiveResource. For representing resources such
    as workers and charging interfaces that are required for a DepotProcess.
    Can be switched on/off using simpyExt.ResourceSwitch.

    Parameters:
    depot: [Depot] instance

    """

    def __init__(self, env, ID, depot, capacity=1):
        super(DepotResource, self).__init__(env, capacity)

        self.env = env
        self.ID = ID
        self.depot = depot

        if globalConstants["general"]["LOG_ATTRIBUTES"]:
            self.logger = DataLogger(env, self, "DEPOTRESOURCE")

    request = BoundClass(DepotResourceRequest)
    release = BoundClass(DepotResourceRelease)

    @property
    def user_count(self):
        """Return a dict that contains the number of current users."""
        user_count = {}
        for user in self.users:
            ID = user.caller.ID
            if ID not in user_count:
                user_count[ID] = 1
            else:
                user_count[ID] += 1
        return user_count


class DepotChargingInterface(DepotResource):
    """
    Parameters:
    max_power: [int or float] in kW

    """

    def __init__(self, env, ID, depot, max_power):
        super(DepotChargingInterface, self).__init__(env, ID, depot, capacity=1)
        self.max_power = max_power
        self._current_power = 0
        self.power_logs_ci = {0: 0}  # Container for power logs at ci

    @property
    def current_power(self):
        return self._current_power

    @current_power.setter
    def current_power(self, value):
        # update depot.total_power
        diff = value - self._current_power
        self.depot.update_power(diff)

        self._current_power = value

        self.power_logs_ci[self.env.now] = self._current_power


class ResourceBreak(BaseDepotProcess):
    """Process used by ResourceSwitch to occupy resources."""

    def __init__(
        self,
        env,
        ID,
        dur,
        required_resources=None,
        resume=False,
        priority=-3,
        recall_priority=-3,
        preempt=True,
        dur_predefined=True,
    ):
        super(ResourceBreak, self).__init__(
            env,
            ID,
            dur,
            required_resources,
            resume,
            priority,
            recall_priority,
            preempt,
            dur_predefined,
        )

    def _action(self):
        try:
            flexprint("Break process started.", env=self.env, switch="res_break")
            yield self.env.timeout(self.dur)
            flexprint("Break process finished.", env=self.env, switch="res_break")

        except simpy.Interrupt:
            # Early reactivation of the resource
            flexprint("Break process interrupted", env=self.env, switch="res_break")
            pass

    @staticmethod
    def estimate_duration(dur, *args, **kwargs):
        return dur


class ResourceSwitch:
    """User that occupies a DepotResource object on a predefined regular
    basis, claiming priority over other users and optionally interrupting them.
    For representing cycles that reduce a resource's capacity such as break
    times for workers.
    Must be activated with 'env.process(run_break_cycle())' before simulation
    start.

    Attributes:
    resource: [DepotResource] or subclass object. Can be None before simulation
        start.
    breaks: [list] of tuples defining phases during which *resource* is
        unavailable. The tuples must contain two values: first is the starting
        time of the unavailability-phase and second is the end time. The times
        are defined as total seconds since 0:00 on the starting day. *breaks*
        is repeated until the simulation end.
        The usual case is defining the rhythm of one day. Example:
            breaks = [(12*3600, 13*3600), (18*3600, 20*3600)]
            -> The resource is unavailable from 12:00 until 13:00 and from
            18:00 until 20:00.
        The definition can include rythms of more than one day. The base-
        timevalue stays at 0:00 on the first day:
            breaks = [(12*3600, 13*3600), (36*3600, 38*3600)]
            -> The resource is unavailable from 12:00 until 13:00 on the first
            day and from 12:00 until 14:00 on the next day.
        An unavailability-phase during the night:
            breaks = [(16.5*3600), (32*3600)]
            -> The resource is only available from 8:00 until 16:30.
        Important: Phases of unavailability must not overlap.
    preempt: [bool] Set to True if a current user's process should be strictly
        interrupted when a break begins. After the end of a break preempted
        requests are not resumed by default. If needed, the request executing
        instance has to implement resuming. Set to False if the process should
        be permitted to finish before the break starts. The end time of the
        break is not postponed, meaning that the break duration might be cut.
    strength: [int or str] of resource.capacity that is unavailable during
        breaks of this user. Set to resource.capacity or 'full' if resource
        should be fully unavailable. Set to between 0 and resource.capacity to
        only partly switch it off.
    priority [int or float]: Priority of break-requests to *resource*. Lower
        number means higher priority. Fixed value of -3. Requests by regular
        users should have a priority > -3 (the default priority for regular
        users is 0). User priority of <= -3 is possible, but can result in
        delayed or never even happening breaks.
    resume: [bool] If True, an interrupted break is resumed for the remaining
        duration.

    """

    def __init__(
        self,
        env,
        ID,
        resource,
        breaks,
        preempt=True,
        strength="full",
        resume=False,
        priority=-3,
    ):
        # TODO temporarily add priority argument
        self.env = env
        self.ID = ID
        self.strength_input = strength
        self._strength = None  # to be set with setter
        self._resource = None  # to be set with setter
        self._breaks = None  # to be set with setter
        self.strength = strength
        self.resource = resource

        self.breaks = breaks
        self.preempt = preempt

        self.priority = priority
        self.resume = resume

    @property
    def breaks(self):
        return self._breaks

    @breaks.setter
    def breaks(self, value):
        if self.check_breaks(value):
            self._breaks = value

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, value):
        """Set self._resource to *value* and update self.strength if value is
        not None. Value must be a [DepotResource] or subclass object
        or None.
        """
        if value is not None:
            if not isinstance(value, DepotResource):
                raise ValueError(
                    "Argument resource to ResourceSwitch must be "
                    "of class DepotResource or subclass."
                )

            self.strength = self.strength_input

        self._resource = value

    @property
    def strength(self):
        """Return None or the current resource.capacity or the custom value
        self._strength.
        """
        if self.resource is None:
            return None
        elif self.strength_input == "full":
            return self.resource.capacity

        return self._strength

    @strength.setter
    def strength(self, value):
        # Check if strength is int or 'full'
        if not isinstance(value, int) and (
            not isinstance(value, str) or value != "full"
        ):
            raise ValueError(
                "'%s' is an invalid input for ResourceSwitch %s " % (value, self.ID)
                + "strength. Must be 'full' "
                "(str) or of type int."
            )

        # Check if 1 < value < self.resource.capacity if int
        if self.resource is not None:
            if isinstance(value, int) and (value > self.resource.capacity or value < 1):
                raise ValueError(
                    'ResourceSwitch "%s" strength must be '
                    "between 1 and the capacity of the related "
                    "resource (=%d)." % (self.ID, self.resource.capacity)
                )

        self.strength_input = value
        self._strength = value

    @staticmethod
    def check_breaks(breaks):
        """Do essential validity checks on ResourceSwitch parameter *break*."""
        if not breaks:
            raise ValueError("ResourceSwitch breaks cannot be empty.")

        for break_no, break_i in enumerate(breaks):
            if (
                not isinstance(break_i, tuple)
                and not isinstance(break_i, list)
                or len(break_i) != 2
            ):
                raise ValueError(
                    "ResourceSwitch break must contain tuples of " "length = 2"
                )
            if break_i[0] < 0 or break_i[1] < 0:
                raise ValueError("ResourceSwitch break times must be nonnegative.")
            if break_i[0] >= break_i[1]:
                raise ValueError(
                    "ResourceSwitch break start time must be earlier "
                    "than break end time"
                )
            if break_no > 0:
                if break_i[0] <= breaks[break_no - 1][1]:
                    raise ValueError(
                        "ResourceSwitch break phases must be listed in "
                        "chronologically ascending order and start times must "
                        "be later than end times of previous phases."
                    )
        return True

    def run_break_cycle(self):
        """Infinitely loop through self.breaks and occupy self.resource
        accordingly.
        """
        # Update strength in case the self.resource.capacity has changed since
        # setting self.resource
        self.strength = self.strength_input

        for cycle_no in itertools.count():
            for break_i in self.breaks:
                # Determine the earliest possible break_start in the future
                day_no = self.env.now // 86400
                if day_no * 86400 + break_i[0] < self.env.now:
                    day_no += 1
                break_start = day_no * 86400 + break_i[0]
                until_break = break_start - self.env.now

                if until_break > 0:
                    # Wait until break
                    yield self.env.timeout(until_break)

                break_duration = break_i[1] - break_i[0]

                for m in range(self.strength):
                    # self.env.process(self.take_break(break_duration))
                    self.take_break(break_duration)

                # Create own timeout event instead of waiting for take_break()
                # because the loop has to continue even if a break is delayed
                yield self.env.timeout(break_duration)

    def take_break(self, duration):
        """Request one slot at self.resource and hold it for *duration*."""
        proc = ResourceBreak(
            self.env,
            "break",
            duration,
            required_resources=[self.resource],
            resume=self.resume,
            priority=self.priority,
            recall_priority=self.priority,
            preempt=self.preempt,
        )

        self.env.process(proc())



# -*- coding: utf-8 -*-
"""
Initialization and hosting of a depot simulation.

"""
import copy
from abc import ABC, abstractmethod
from collections import namedtuple

import simpy
from eflips.helperFunctions import load_json, save_json, set_by_path

import eflips


class DepotHost:
    """Host for one depot and its configurator in a simulation. Needs to be
    operated by a SimulationHost or a similar procedure.
    Attributes depot, configurator and evaluation are stable references.

    Parameters:
    simulation_host: [SimulationHost]

    Attributes:
    depot: related [Depot]
    configurator: [DepotConfigurator] of related depot
    evaluation: [DepotEvaluation] of related depot
    """

    def __init__(self, env, simulation_host):
        self.env = env
        self.simulation_host = simulation_host

        self.evaluation = eflips.depot.DepotEvaluation(self)
        self.configurator = eflips.depot.DepotConfigurator(self.env)
        self.depot = self.configurator.depot

    def load_and_complete_template(self, filename_template):
        """Load depot template, validate and complete it."""
        success, errormsg = self.configurator.load(filename_template)
        if not success:
            raise ValueError("Error while loading template: " + errormsg)

        success, errormsg = self.configurator.complete()
        if not success:
            raise ValueError(errormsg)


class SimulationHost:
    """Wrapper for one or more depots in one simulation run.
    Provides utilities to properly setup a simulation:
    - load eflips settings
    - Create simpy.env, a timetable and a vehicleGenerator
    - Create one or more DepotHost instances and related objects
    - Measure time
    - Run the simulation

    Makes it possible to start the simulation from an outside script.

    Parameters:
    to_simulate: [list] of Depotinput objects. Specifies the number of
        depots to simulate, their template and if a GUI should be shown. Based
        on this input, DepotHost instances are created.

        The parallel simulation of multiple depots has not been tested yet,
        therefore to_simulate must have only one entry.

    Attributes:
    tictoc: [eflips.helperFunctions.Tictoc] measures execution time
    env: [simpy.Environment]
    depot_hosts: [list] of DepotHost instances
    timetable: [eflips.depot.standalone.Timetable]
    vg: [eflips.depot.standalone.VehicleGenerator]
    """

    def __init__(
        self, to_simulate, run_progressbar=False, print_timestamps=True, tictocname=""
    ):
        self.to_simulate = to_simulate

        self.tictoc = eflips.helperFunctions.Tictoc(print_timestamps, tictocname)
        self.tictoc.tic()
        self.run_progressbar = run_progressbar

        self.env = simpy.Environment()

        self.vg = eflips.depot.VehicleGenerator(self.env)

        self.gc = eflips.globalConstants

        self.filename_timetable = None
        self.timetable = None

        # Instantiate depot host(s) with empty depots
        self.depot_hosts = [DepotHost(self.env, self) for _ in to_simulate]
        self.depots = [dh.depot for dh in self.depot_hosts]

    @property
    def filename_eflips_settings(self):
        return self.gc["FILENAME_SETTINGS"] if "FILENAME_SETTINGS" in self.gc else None

    def standard_setup(self, filename_eflips_settings, filename_timetable):
        """Standard simulation setup without gui."""
        self.load_eflips_settings(filename_eflips_settings)
        self.load_timetable(filename_timetable)
        for dh, di in zip(self.depot_hosts, self.to_simulate):
            dh.load_and_complete_template(di.filename_template)
        self.complete()

    @staticmethod
    def load_eflips_settings(filename):
        """Load eflips.globalConstants. Validate and complete it (required for
        the depot simulation).
        """
        eflips.load_settings(filename)
        eflips.depot.settings_config.check_gc_validity()
        eflips.depot.settings_config.complete_gc()

    def load_timetable(self, filename):
        """Load timetable data from excel and init a timetable.
        load_eflips_settings() must be called before this.
        """
        timetabledata = eflips.depot.standalone.timetabledata_from_excel(filename)
        self.init_timetable(timetabledata)

    def init_timetable(self, timetabledata):
        """Use timetabledata to init a timetable."""
        self.timetable = eflips.depot.standalone.timetable_from_timetabledata(
            self.env, timetabledata
        )
        self.filename_timetable = timetabledata.filename.replace(".xlsx", "")

    def complete(self):
        """Complete the depot configuration phase. Must be called after
        loading a depot template/ configurating it and before simulation start.
        """
        for depot in self.depots:
            depot.timetable = self.timetable

        for depot_host in self.depot_hosts:
            depot_host.evaluation.complete()

    def run(self):
        """Run the simulation. All depot configurations have to be complete."""
        self.vg.run(self.depots)
        self.env.process(self.timetable.run(self.depots))

        if self.run_progressbar:
            self.env.process(
                eflips.helperFunctions.progressbar(
                    self.env,
                    eflips.settings.globalConstants["general"]["SIMULATION_TIME"],
                    step=10,
                    step_unit="%",
                )
            )

        self.tictoc.toc("list")  # mark the end of the configuration phase

        # Run env
        self.env.run(
            until=eflips.settings.globalConstants["general"]["SIMULATION_TIME"]
        )

        self.tictoc.toc("list")  # mark the end of the simulation phase
        if self.tictoc.print_timestamps:
            self.tictoc.print_toclist("interval")
            self.tictoc.print_toclist("cumulative")


Depotinput = namedtuple("Depotinput", ["filename_template", "show_gui"])
"""Container for parameters for SimulationHost"""


def create_alternatives(basefilename, keypath, values):
    """Create one or more json depot templates that vary by one attribute.
    Parameters:
    basefilename: [str] including path, excluding extension
    keypath: [list] of access keys
    values: [list] of value alternatives
    """
    basedata = load_json(basefilename)
    for i, value in enumerate(values):
        data = copy.deepcopy(basedata)
        set_by_path(data, keypath, value)
        save_json(data, basefilename + "_" + str(i))
        print("Saved as %s" % basefilename + "_" + str(i))


class BaseMultipleSimulationHost(ABC):
    """Base utilities for consecutively running multiple depot simulations.

    See usage demo in run_multiple_test.py

    history: [list] of tuples containing the filenames of the inputs of each
    simulation run.
    """

    def __init__(
        self,
        basefilename_eflips_settings,
        basefilename_timetable,
        basefilename_template,
        print_timestamps_total=True,
        tictocname_total="",
        print_timestamps_each=True,
        tictocname_each="",
    ):
        self.tictoc = eflips.Tictoc(print_timestamps_total, tictocname_total)
        self.tictoc.tic()
        self.print_timestamps_each = print_timestamps_each
        self.tictocname_each = tictocname_each

        self.basefilename_eflips_settings = basefilename_eflips_settings
        self.basefilename_timetable = basefilename_timetable
        self.basefilename_template = basefilename_template
        self.history = []

    def run(self):
        """Call to execute."""
        fn_es = self.basefilename_eflips_settings
        fn_ti = self.basefilename_timetable
        fn_te = self.basefilename_template
        done = False

        while not done:
            print()
            print()
            eflips.settings.reset_settings()

            simulation_host = self.simulate(fn_es, fn_ti, fn_te)

            self.evaluate(simulation_host.depot_hosts[0])

            done, fn_es, fn_ti, fn_te = self.next(
                simulation_host.depot_hosts[0], fn_es, fn_ti, fn_te
            )

        self.tictoc.toc("list")
        if self.tictoc.print_timestamps:
            self.tictoc.print_toclist("cumulative")

    def simulate(self, fn_es, fn_ti, fn_te):
        """Execute one simulation and return its SimulationHost instance."""
        simulation_host = eflips.depot.SimulationHost(
            [eflips.depot.Depotinput(filename_template=fn_te, show_gui=False)],
            run_progressbar=False,
            print_timestamps=self.print_timestamps_each,
            tictocname=self.tictocname_each,
        )
        simulation_host.standard_setup(fn_es, fn_ti)
        print(
            "Simulating template %s."
            % simulation_host.depot_hosts[0].configurator.templatename
        )
        simulation_host.run()
        print("Completed simulation.")
        self.history.append((fn_es, fn_ti, fn_te))
        return simulation_host

    def evaluate(self, depotsim):
        """Do evaluation after a simulation run."""

    @abstractmethod
    def stop_criterion(self, depotsim):
        """Return True if simulating should be stopped, else False."""
        return True

    @abstractmethod
    def next(self, depotsim, fn_es, fn_ti, fn_te):
        """
        Call self.stop_criterion to decide on further actions. For alternating,
        create a new scenario to simulate based on the results of the previous
        run.
        Return a tuple of
            - return value of self.stop_criterion,
            --- fn_es, fn_ti, fn_te that may each point to another file than
                passed as parameter, possibly newly created.
        """
        # Decide if continue or stop
        if not self.stop_criterion(depotsim):
            # Stop
            return True, None, None, None
        else:
            # Alternate
            # ...
            return False, fn_es, fn_ti, fn_te


class SimulationStop(simpy.Event):
    """Event for immediately terminating the simulation."""

    def __init__(self, env):
        super(SimulationStop, self).__init__(env)
        self.callbacks.append(simpy.core.StopSimulation.callback)

    def succeed(self, value=None):
        """Same as simpy.Event.succeed() but with URGENT priority.
        Use value to e.g. supply a reason.
        """
        if self._value is not simpy.events.PENDING:
            raise RuntimeError("%s has already been triggered" % self)

        self._ok = True
        self._value = value
        self.env.termination_reason = self
        self.env.schedule(self, simpy.events.URGENT)
        return self
